function [X_total,synth_samples] = ADASYN(X_min,X_maj,beta,prior_opt,sim_measure,K,K_tilde,nn_sl,sim_sl,nn_s,sim_s)
% This function applies the ADASYN_ORIG technique for binary data and generates
% synthetic samples. 

%Inputs
%   X_min: original binary matrix of MINORITY class samples
%   X_maj: original binary matrix of MAJORITY class samples
%
%   beta: controls the amount of synth samples generated 
%         (G = (m_l - m_s)*beta, m_l and m_s are the amount of
%                               maj and min examples respectively.
%
%   prior_opt: This deals with the handling of 0-1 and 1-0 case when
%              comparing a certain original sample with a specific nearest 
%              neighbour under consideration.
%              'FlipCoin': in 50% of cases, the synthetic sample will have 
%              a 1 at the position of 0-1 or 1-0 occurence.
%
%              'Prior': The value of the synthetic sample at the position
%              of a 0-1 or 1-0 occurence is determined by the prior
%              distribution in the colomn corresponding to this position.
%
%              'Reverse_Prior': Analog as above, nl. prior = 1-prior of
%              the 'Prior'-case.
%
%   sim_measure: measure that determines which point is closest in 
%               computing the nearest neighbours
%               'Jaccard' a/(a+b+c)
%               'Cosine' 
%
%   K: number of nearest neighbours computed (from minclass).
%  
%   K_tilde: nr of nearest neighbours computed (from min & maj)
%   
%   nn_sl,sim_sl (optional): result of nearest_neighbours
%                           (X_min,[X_min;X_maj],K_tilde_max,sim_measure);
%
%   nn_s,sim_s (optional): result of nearest_neighbours
%                           (X_min,X_min,K_max,sim_measure);

%Outputs
%   X_total: matrix containing the original X_min and the synthetic samples.
%   synth_samples: matrix containing the synthetic samples

%nr of minority and majority examples
m_s = size(X_min,1);
m_l = size(X_maj,1);

%total amount of synth samples to be generated
G = ceil((m_l - m_s)*beta);
if(beta == 0)
    synth_samples = [];
    X_total = X_min;
    return
end

%Calculate K_tilde nn from {minclass,majclass}
if( ~exist('nn_sl','var') || isempty(nn_sl)  || ~exist('sim_sl','var') || isempty(sim_sl))
    [nn_sl,sim_sl] = nearest_neighbours(X_min,[X_min;X_maj],K_tilde,sim_measure,1000,100000);
else
    nn_sl = nn_sl(:,1:K_tilde);
    sim_sl = sim_sl(:,1:K_tilde);
end
%determine Delta_i, weights r_i and normalized weights r_norm_i
Delta = sum((nn_sl > m_s | sim_sl == 0),2);
r = Delta/K_tilde;
r_norm = r/sum(r);
if(r_norm == 0)
    r_norm = repmat(1/m_s,m_s,1);
end
%Determine nr of synth_samples for each x_i (g_i = r_norm_i*G)
g = ceil(r_norm*G);

%Calculate K nn for each point in X_min (neighbours chosen from X_min)
if( ~exist('nn_s','var') || isempty(nn_s)  || ~exist('sim_s','var') || isempty(sim_s))
    [nn_s,sim_s] = nearest_neighbours(X_min,X_min,K,sim_measure,1000,100000);
else
    nn_s = nn_s(:,1:K);
    sim_s = sim_s(:,1:K);
end

%Find priors for each colummn in X_min
switch prior_opt
    case 'FlipCoin'
        prior = 0.5*ones(1,size(X_min,2));
    case 'Prior'
        prior = full(sum(X_min,1))/m_s;
    case 'Reverse_Prior'
        prior = full(sum(X_min,1))/m_s;
        prior = 1-prior;
end

%initialize rows, cols and values which are for storing the synthetic
%samples   (rows,cols,values) = find(synth_samples)
rows = zeros(1,2*ceil(G/m_s)*nnz(X_min));
cols = zeros(1,2*ceil(G/m_s)*nnz(X_min));
values = zeros(1,2*ceil(G/m_s)*nnz(X_min));
rcv_counter = 1; %position we should store new elem in rows,cols & values

%Determine the synthetic samples
for i = 1:m_s
    %tic;
    %We generate g_i synth samples for the i-th original sample
    if(g(i) == 0)
        continue
    end
    %duplicate X_min(i,:) and access nearest neighbours for this point
    x_i = X_min(i,:);
    nn_x_i = nn_s(i,:);% k nearest neighbours positions for point x_i
    if(~isempty(find(sim_s(i,:)~=0, 1)))
        nn_x_i = nn_x_i(sim_s(i,:)~=0);%remove nn that have sim = 0
    else    
        nn_x_i = i;% nn = x_i      
    end
    g_i_nn_x_i = nn_x_i( randi(length(nn_x_i),1,g(i)) );% g_i nearest neigbours for point x_i

    %only compute x_i_rep + X_min(g_i_nn_x_i,:) for unique nn's
    [unique_g_i_nn,~,pos] = unique(g_i_nn_x_i);
    x_i_rep = repvec(x_i,length(unique_g_i_nn));
    x_tot = x_i_rep + X_min(unique_g_i_nn,:);
    %pos gives position of each g_i_nn_x_i in unique g_i_nn
    x_tot = x_tot(pos,:);
    [r_tot,c_tot,v_tot] = find(x_tot);
    
    %for each nn of x_i, determine where 0-1 or 1-0 occurs when comparing
    %this nn with sample x. Apply the prior_opt next.
    pos_nonmatch = find(v_tot == 1);
    r_nonmatch = r_tot(pos_nonmatch);c_nonmatch = c_tot(pos_nonmatch);
    v_nonmatch = v_tot(pos_nonmatch);
    if(~isempty(r_nonmatch))
        v_nonmatch = rand(size(v_nonmatch));
        v_nonmatch = v_nonmatch <= reshape(prior(c_nonmatch),size(v_nonmatch));
    end
    
    %for each nn of x_i, determine where 1-1 occurs when comparing this nn
    %with sample x_i.
    pos_match = find(v_tot == 2);
    r_match = r_tot(pos_match);c_match = c_tot(pos_match);
    v_match = ones(size(r_match));
    
    %construct final synth_samples for x_i
    %(synth_samples x_i = find(r,c,v))
    r = [r_nonmatch(:); r_match(:)];c = [c_nonmatch(:); c_match(:)];v = [v_nonmatch(:); v_match(:)];
    
    %update rows,cols,values and rcv_counter
    r = r + sum(g(1:i-1));
    nr_nonzeros = length(r);
    rows(rcv_counter:rcv_counter+nr_nonzeros-1) = r;
    cols(rcv_counter:rcv_counter+nr_nonzeros-1) = c;
    values(rcv_counter:rcv_counter+nr_nonzeros-1) = v;
    rcv_counter = rcv_counter + nr_nonzeros;
    %toc
end

%delete unnecessary rows,cols and values
rows = rows(rows~=0);
cols = cols(rows~=0);
values = values(rows~=0);

%Construct synth_samples_matrix (too large at this stage)
synth_samples = sparse(rows,cols,values,sum(g),size(X_min,2)); 

%Randomly remove synth points until total nr of synth samples equals G
while(size(synth_samples,1) ~= G)
    pos = randi(size(synth_samples,1),size(synth_samples,1)-G,1);
    synth_samples = removerows(synth_samples,'ind',pos);
end

X_total = [X_min;synth_samples];

end

